import{r as o,qs as i,qt as r,bp as s,qu as t,qv as v,qw as a}from"./index-fPNPEmeW.js";import{u as c}from"./useTranslation-_h2To1XB.js";function l({children:n}){const{i18n:p}=c(),e=p.dir(),x=o.useMemo(()=>i({...r,direction:e,shadows:{...r.shadows,selected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-500), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverSelected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-400), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverUnselected:"inset 0px 0px 0px 2px var(--invoke-colors-invokeBlue-300), inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-800)",selectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-300), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)",hoverSelectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-200), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)"}}),[e]);return o.useEffect(()=>{document.body.dir=e},[e]),s.jsx(t,{theme:x,toastOptions:v,children:s.jsx(a,{children:n})})}const d=o.memo(l);export{d as default};
